/*
 * Copyright (c) 2003, 2004 Henri Sivonen
 *
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 */

package fi.iki.hsivonen.xml;
import java.util.Locale;
import java.util.NoSuchElementException;
import java.util.StringTokenizer;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import fi.iki.hsivonen.gnu.xml.pipeline.DomConsumer;
import fi.iki.hsivonen.schemas.dtd.DTDCatalog;

/**
 * A collection of utility methods for working with the DOM.
 *
 * @author hsivonen
 */
public class DOMUtils {

    /**
     * Finds the first occurrence of an element in the subtree rooted at 
     * <code>node</code>
     *
     * @param node the root of the subtree to search
     * @param namespace the namespace URI of the element being seached
     * @param localName the local name of the element being seached
     *
     * @return the first occurrence of the named element or <code>null</code>
     *         if not found
     */
    public static final Element findElement(
        Node node,
        String namespace,
        String localName) {
        Node current = node;
        Node next;
        for (;;) {
            switch (current.getNodeType()) {
                case Node.ELEMENT_NODE :
                    if (localName.equals(current.getLocalName())
                        && namespace.equals(current.getNamespaceURI())) {
                        return (Element)current;
                    }
                    // fall through
                case Node.DOCUMENT_FRAGMENT_NODE :
                case Node.DOCUMENT_NODE :
                    if ((next = current.getFirstChild()) != null) {
                        current = next;
                        continue;
                    }
            }
            for (;;) {
                if ((next = current.getNextSibling()) != null) {
                    current = next;
                    break;
                }
                current = current.getParentNode();
                if (current == node)
                    return null;
            }
        }
    }

    /**
     * Finds an element of that has an attribute called <code>id</code> which 
     * has the given value and is not in a namespace. The IDness of the 
     * attribute is based on the attribute name--not on the DTD.
     * 
     * @param node the root of the subtree to search
     * @param id the value of the id attribute
     * @return the first element that has the specified attribute
     */
    public static final Element getElementById(Node node, String id) {
        Node current = node;
        Node next;
        for (;;) {
            switch (current.getNodeType()) {
                case Node.ELEMENT_NODE :
                    Element elt = (Element)current;
                    if (id.equals(elt.getAttribute("id"))) {
                        return elt;
                    }
                    // fall through
                case Node.DOCUMENT_FRAGMENT_NODE :
                case Node.DOCUMENT_NODE :
                    if ((next = current.getFirstChild()) != null) {
                        current = next;
                        continue;
                    }
            }
            for (;;) {
                if ((next = current.getNextSibling()) != null) {
                    current = next;
                    break;
                }
                current = current.getParentNode();
                if (current == node)
                    return null;
            }
        }
    }

    /**
     * Returns the white space-normalized text content of the subtree rooted at 
     * <code>node</code>.
     *
     * @param node the subtree
     * @return the white space-normalized text content
     */
    public static final String textContent(Node node) {
        StringBuffer buf = new StringBuffer();
        boolean lastIsWhitespace = true;
        Node current = node;
        Node next;
        for (;;) {
            switch (current.getNodeType()) {
                case Node.TEXT_NODE :
                case Node.CDATA_SECTION_NODE :
                    String text = current.getNodeValue();
                    for (int i = 0; i < text.length(); i++) {
                        char c = text.charAt(i);
                        if (c == ' ' || c == '\t' || c == '\n') {
                            if (!lastIsWhitespace) {
                                buf.append(' ');
                                lastIsWhitespace = true;
                            }
                        } else {
                            buf.append(c);
                            lastIsWhitespace = false;
                        }
                    }
                    break;
                case Node.ELEMENT_NODE :
                case Node.DOCUMENT_FRAGMENT_NODE :
                case Node.DOCUMENT_NODE :
                    if ((next = current.getFirstChild()) != null) {
                        current = next;
                        continue;
                    }
                    break;
            }
            for (;;) {
                if ((next = current.getNextSibling()) != null) {
                    current = next;
                    break;
                }
                current = current.getParentNode();
                if (current == node) {
                    if (buf.charAt(buf.length() - 1) == ' ') {
                        buf.deleteCharAt(buf.length() - 1);
                    }
                    return buf.toString();
                }
            }
        }
    }

    /**
     * Instantiates a <code>DocumentBuilder</code> which is non-validating, 
     * is namespaces aware, expands entities using the local 
     * <code>DTDCatalog</code>, doesn't make arbitrary network connections, 
     * coalesces CDATA sections and ignores comments.
     *
     * @return a configured <code>DocumentBuilder</code>
     */
    public static final DocumentBuilder newNonvalidatingDocumentBuilder() {
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        factory.setNamespaceAware(true);
        factory.setExpandEntityReferences(true);
        factory.setCoalescing(true);
        factory.setIgnoringComments(true);
        factory.setIgnoringElementContentWhitespace(false);
        factory.setValidating(false);
        try {
            DocumentBuilder builder = factory.newDocumentBuilder();
            builder.setEntityResolver(DTDCatalog.getInstance());
            return builder;
        } catch (ParserConfigurationException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * Checks whether a node is a text or CDATA node consisting of white 
     * space only.
     *
     * @param node the node to examine
     *
     * @return <code>true</code> if it is a white space node and 
     * <code>false</code> otherwise
     */
    public static final boolean isWhiteSpace(Node node) {
        String value = node.getNodeValue();
        for (int i = 0; i < value.length(); i++) {
            char c = value.charAt(i);
            if (!(c == ' ' || c == '\t' || c == '\n')) {
                return false;
            }
        }
        return true;
    }

    public static final String language(Node node) {
        String lang;
        for (;;) {
            if (node == null) {
                return "";
            }
            switch (node.getNodeType()) {
                case Node.ELEMENT_NODE :
                    if ((lang =
                        ((Element)node).getAttributeNS(
                            "http://www.w3.org/XML/1998/namespace",
                            "lang"))
                        != null) {
                        return lang;
                    } else {
                        node = node.getParentNode();
                    }
                    break;
                case Node.CDATA_SECTION_NODE :
                case Node.TEXT_NODE :
                    node = node.getParentNode();
                    break;
                default :
                    return "";
            }
        }
    }

    public static final Locale languageAsLocale(Node node) {
        String lang = language(node);
        if ("".equals(lang)) {
            return null;
        }
        StringTokenizer tok = new StringTokenizer(lang, "-");
        String language = null;
        String country = null;
        try {
            language = tok.nextToken();
            country = tok.nextToken();
        } catch (NoSuchElementException e) {
        }
        if (language == null) {
            return null;
        }
        if (country == null) {
            return new Locale(language);
        } else {
            return new Locale(language, country);
        }
    }

    public static final Element findBody(Document doc) {
        for (Node n = doc.getDocumentElement().getLastChild();
            n != null;
            n = n.getPreviousSibling()) {
            if ("body".equals(n.getLocalName())
                && "http://www.w3.org/1999/xhtml".equals(n.getNamespaceURI())) {
                return (Element)n;
            }
        }
        return null;
    }

    public static final void importChildrenBefore(
        Node from,
        Node to,
        Node ref) {
        Document doc = to.getOwnerDocument();
        for (Node curr = from.getLastChild();
            curr != null;
            curr = curr.getPreviousSibling()) {
            ref = to.insertBefore(doc.importNode(curr, true), ref);
        }
    }

    /**
     *
     */
    public static DomConsumer newDomConsumer() {
        try {
            return new DomConsumer(
                DocumentBuilderFactory
                    .newInstance()
                    .newDocumentBuilder()
                    .newDocument()
                    .getClass());
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }
}
