/*
 * Copyright (c) 2003, 2004 Jussi Koskela
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */
package fi.iki.hsivonen.webcache.dependency;

import java.net.URL;

public class CacheInvalidator 
{
    /** Do we print the debug information */
    private static boolean DEBUG = false;

    /** The instance of CacheInvalidator */
    private static CacheInvalidator theInstance;

    /** The coordinator thread */
    private CoordinatorThread coordinator;

    /**
     * Class constructor. CacheInvalidator is singleton, so the constructor 
     * is declared private. Use getInstance-method to get the instance of 
     * CacheInvalidator.
     */
    protected CacheInvalidator() {
        DEBUG = (System.getProperty("DebugInvalidator") != null);
        coordinator = new CoordinatorThread();
        coordinator.setDaemon(true);
        coordinator.start();
    }

    /**
     * Prints the given debug string.
     * @param debug the debug string
     */
    public static void debug(String debug) {
        if (DEBUG) {
            System.out.println(debug);
        }
    }

    /**
     * Returns the instance of CacheInvalidator.
     * @return the instance of CacheInvalidator
     */
    public static final synchronized CacheInvalidator getInstance() {
        if (theInstance == null) {
            theInstance = new CacheInvalidator();
        }
        return theInstance;
    }

    /**
     * Starts the dependency tracker.
     * @param servers the URLS for the cache servers
     */
    public void start(URL[] servers) {
        for (int i = 0; i < servers.length; i++) {
            startServer(servers[i]);
        }
    }

    /**
     * Starts sending invalidations to the given cache server.
     * @param server the URL of the cache server
     */
    public void startServer(URL server) {
        coordinator.startServer(server);
    }
    
    /**
     * Stops sending invalidations to the given cache server.
     * @param server the url of the cache server
     */
    public void stopServer(URL server) {
        coordinator.stopServer(server);
    }
    
    /**
     * Shuts down the cache invalidator.
     */
    public void shutdown() {
        coordinator.kill();
    }                    

    /**
     * Invalidates the content which depends on the given key.
     * @param key the key to invalidate
     */
    public void invalidate(String key) {
        coordinator.invalidate(key);
        debug("Invalidating " + key);
    }
}
